/*
	File:		SqueezeToLetterbox.cpp
	
        based on the implementation of Adjust Colors effect in the iMovie SDK from iMovie Engineering team

        Squeeze to Letterbox iMovie plug in
        
        Written to perform a conversion on all frames of an iMovie clip from 4x3 to 16x9 with letterbox.  
        
        Background: Some video cameras are capable of shooting in 16x9 (widescreen) format rather than 
        the standard TV format of 4x3.  Such cameras usually stretch the data to the full 4x3 rectangle 
        in order to maximize the vertical resolution of the stored video.  
        
        When handling this data, iMovie performs no interpretation of the data; i.e., iMovie presents 
        the vertically 16x9 data in a 4x3 rectangle.  This effects filter is designed to efficiently 
        and cleanly squeeze that data into the correct aspect ratio, 16x9, accompanied by the usual 
        letterbox bars.  Such a filter exists elsewhere, but this author thought it was silly 
        to pay upwards of $25 to buy a package just to perform such a trivial operation.  
        
****************************************************************
License:

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

The GNU General Public License is available by:
Writing to the Free Software Foundation, Inc., 
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
or accessing http://www.fsf.org/copyleft/gpl.html

****************************************************************

	Change History (most recent first):
        
        2003 July 19: Creation by Dean E. Dauger (DED)
            the most important changes are in DoSqueezeToLetterboxEffect, including a vectorized version
        
*/

#include <Stdio.h>
#include <String.h>
#include <Math.h>
#include <Carbon/Carbon.h>
#include "HYPluginInterface.h"

#define	kSliderLabelStringResID 128
#define kSliderMinMaxLabelResID	129

#define REMEMBER_LAST_SETTINGS 0

#define prox (params->procs)

enum {
	kSqueezeToLetterbox = 1
};

#define	bicubicInterpolation		1
#define linearInterpolation		0

typedef struct _EffectData {
	long frameCount;			
	long effectSourceCount;
	float sliderValues[MAX_FILTER_SLIDERS];
        short interlaced /* vs. progressive */;
        short interpolation; 
        short useAltiVec /*, if available, vs. no altiVec */; 
        short unusedS; 
	long fadeInFrames;
	long fadeOutFrames;
	GWorldPtr	effectGWorld;
} EffectData;

/**
	globals.
**/
static HYPluginFilterInfo	gsLastFilterSettings;
static Boolean				gsSettingsValid = false;

/**-----------------------------------------------------------------------------
 **
 **	Functions
 **
 **/

static HYPluginResult EffectGetInfo ( short which, HYPluginParams *params );
static HYPluginResult EffectBegin ( short which, HYPluginParams *params );
static HYPluginResult EffectDrive ( short which, HYPluginParams *params );
static HYPluginResult EffectEnd ( short which, HYPluginParams *params );
static HYPluginResult MakeEffectPlaceHolder ( short which, HYPluginParams *params );

/**-----------------------------------------------------------------------------
 **
 **	Functions
 **
 **/

//#include <altivec.h>
#include <Carbon/Carbon.h>
short IsAltiVec()
	{long processorAttributes; //gestalt.h
	short out=0; 
	OSErr err = Gestalt(gestaltPowerPCProcessorFeatures, &processorAttributes);
	if (noErr==err) {
		out = 
			(processorAttributes&(1L<<gestaltPowerPCHasVectorInstructions))?1:0;
		}
	else /*SysBeep(1)*/; 
	return out; 
	}


// procedures
int main ( short which, HYPluginParams *params )
{
	HYPluginResult result = kPluginOK;
	
	if ( !params ) {
		return kPluginBadParams;
	}
	if ( params->interfaceVersion < kPluginInterfaceVersion )
		return kPluginAppIsTooOld;
	// the app being newer is okay for now; assumed to be backward compatible

	switch ( params->selector ) {
		case kPlugGetInfo:
			result = EffectGetInfo ( which, params );
			break;
		case kPlugInitialize:
			result = EffectBegin ( which, params );
			break;
		case kPlugFilePlaceHolder:
			result = MakeEffectPlaceHolder ( which, params );
			break;
		case kPlugDoFrame:
			result = EffectDrive ( which, params );
			break;
		case kPlugTerminate:
			result = EffectEnd ( which, params );
			break;
		case kPlugCustomButtonClick:
			// does nothing in this example.  developers
			// can implement custom behavior here.
			printf("custom button clicked\n");
			break;
		default:
			break;
	}

	return result;
}


/*
	Load up the getinfo field for filter plugins.
*/
static void	SetFilterParamInfo ( short which, HYPluginParams *params )
{
	long 	sliderCount;
	Str255	name;
	short idx;
	float	defaultValues[MAX_FILTER_SLIDERS];
	
	switch ( which ) {

		default:
		case kSqueezeToLetterbox:
			defaultValues[0] = 1.0; 
			defaultValues[1] = 1.0;
                        defaultValues[2] = IsAltiVec()?1.0:0;
			sliderCount = IsAltiVec()?3:2;
			break;
	}

	// for all these filters, which are 1-1, use the selectedFrames field to fill out the
	// takeFromBefore, etc
	params->getInfo.takeFromClipBefore = 0;
	params->getInfo.takeFromClipAfter = params->getInfo2.selectedFrames;
	params->getInfo.framesToBeRendered = params->getInfo2.selectedFrames;
	
	for ( idx = 0; idx < sliderCount; ++idx ) {

		params->getInfo2.filterSliders[idx].defaultValue = gsSettingsValid ? gsLastFilterSettings.sliderValues[idx] : defaultValues[idx];
		params->getInfo2.filterSliders[idx].defaultValue = gsSettingsValid ? gsLastFilterSettings.sliderValues[idx] : defaultValues[idx];

		// Slider names
		GetIndString ( name, kSliderLabelStringResID, idx + 1 );
		BlockMoveData ( name + 1, params->getInfo2.filterSliders[idx].label, name[0]);
		params->getInfo2.filterSliders[idx].label[ name [0] ] = 0;
		
		// Slider min and max labels: resource arranged min1 max1 min2 max2 min3 max3 ...
		GetIndString ( name, kSliderMinMaxLabelResID, 2 * idx + 1 );
		BlockMoveData ( name + 1, params->getInfo2.filterSliders[idx].minLabel, name[0]);
		params->getInfo2.filterSliders[idx].minLabel[ name [0] ] = 0;
		GetIndString ( name, kSliderMinMaxLabelResID, 2 * idx + 2 );
		BlockMoveData ( name + 1, params->getInfo2.filterSliders[idx].maxLabel, name[0] );
		params->getInfo2.filterSliders[idx].maxLabel[ name [0] ] = 0;

	}
	
	prox->strcpy(params->getInfo3.buttonTitle, "Bar Color");
	
}


static HYPluginResult EffectGetInfo ( short which, HYPluginParams *params )
{
	HYPluginResult result = kPluginOK;
	long frameCount;
	
	switch ( which ) {
	
		case kSqueezeToLetterbox:
		default:
			params->getInfo.pluginType = kPluginFilterType;
			frameCount = 0;		// filters operate on selection not a chosen set of frames.
			params->getInfo.framesToBeRendered = frameCount;
			params->getInfo.takeFromClipBefore = 0;
			params->getInfo.takeFromClipAfter = frameCount;
			params->getInfo.minSpeedFrames = 0;
			params->getInfo.maxSpeedFrames = 0;
			params->getInfo.mustTakeFromSingleClips = false;
			// file and custom button flags set to show UI.  They do nothing in this example.
			params->getInfo.pluginFlags |=0; //kPluginChoiceCustomButton;
    // (kFilterChoiceFadeIn | kFilterChoiceFadeOut | kPluginChoiceFileButton);
			
			params->getInfo2.baseMemoryNeeded = 50000;
			
			SetFilterParamInfo ( which, params );
			
			break;

	}
			
	return result;
}

static HYPluginResult EffectBegin ( short which, HYPluginParams *params )
{
	HYPluginResult result = kPluginOK;
	short 							frameWidth, frameHeight, index;
	EffectData 						*data;
	HYRefDigitalVideoStandard		theStandard;
	long							totalFrames;
	OSErr							status = noErr;
	HYPluginFilterInfo				*filterInfo = NULL;

	if ( !params->clientData ) {
		params->clientData = (EffectData *)prox->NewPtrClear ( sizeof(EffectData) );
		if ( !params->clientData )
			return kPluginOutOfMemory;
	}
	data = (EffectData *)params->clientData;

	switch ( which ) {
		case kSqueezeToLetterbox:
		default:
			data->effectSourceCount = 1;
			break;
	}

	frameWidth = params->frameRect.right - params->frameRect.left;
	frameHeight = params->frameRect.bottom - params->frameRect.top;
	
	params->procs->GetMovieInfo ( &theStandard, &frameWidth, &frameHeight, &totalFrames );
	
	result = prox->GetFilterInfo ( params->instance, &filterInfo );
	data->fadeInFrames = filterInfo->fadeInFrames;
	data->fadeOutFrames = filterInfo->fadeOutFrames;
	for ( index = 0; index < MAX_FILTER_SLIDERS; index++ )
		data->sliderValues[index] = filterInfo->sliderValues[index];
        
        data->interlaced = data->sliderValues[0]<0.5?0:1; 
        data->interpolation = data->sliderValues[1]<0.5?linearInterpolation:bicubicInterpolation; 
        data->useAltiVec = data->sliderValues[2]<0.5?0:IsAltiVec()?1:0; 
        
	if ( filterInfo != NULL )
		prox->ReleaseFilterInfo ( params->instance, filterInfo );
		
	if ( data->fadeInFrames != 0 && data->fadeOutFrames != 0 ) {
		if ( data->fadeInFrames > params->frameCount / 2 )
			data->fadeInFrames = params->frameCount / 2;
		if ( data->fadeOutFrames > params->frameCount / 2 )
			data->fadeOutFrames = params->frameCount / 2;	
	} else if ( data->fadeInFrames != 0 && data->fadeInFrames > params->frameCount ) {
		data->fadeInFrames = params->frameCount;
	} else if ( data->fadeOutFrames != 0 && data->fadeOutFrames > params->frameCount ) {
		data->fadeOutFrames = params->frameCount;
	}
	
	data->effectGWorld = NULL;
	
	if ( status != noErr ) {
		if ( result == kPluginOK ) {
			result = kPluginFailed;
		}
	}
		
	if ( result != kPluginOK ) {
	}
	
	return result;
}

static HYPluginResult DoSqueezeToLetterboxEffect ( GWorldPtr sourceFrame, HYPluginParams *params )
{
	HYPluginResult				result = kPluginOK;
	OSErr						status = noErr;
	PixMapHandle				srcPixMapHandle = NULL, dstPixMapHandle = NULL, framePixMapHandle = NULL;
	GWorldFlags					srcPixFlags, dstPixFlags;
	// effect engine
	unsigned char 				*srcBaseAddr, *dstBaseAddr;
	long 						srcRowBytes, dstRowBytes;
	long						row, rowMax, colMax;
	Rect						dstRect, srcRect;
        short interlaced=1, interpolation=1, useAltiVec=0; 
	
	EffectData *data = (EffectData *)params->clientData;	
        
        if (!data->interlaced) interlaced=0;
        if (!data->interpolation) interpolation=0;
        if (data->useAltiVec) useAltiVec=1;

	GetPortBounds(sourceFrame, &srcRect);
	OffsetRect ( &srcRect, - srcRect.left, -srcRect.top );
	
	srcPixMapHandle = GetGWorldPixMap( sourceFrame );
	srcPixFlags = GetPixelsState ( srcPixMapHandle );
	LockPixels( srcPixMapHandle );
	
	// Don't store this huge piece of cdap in the instance, or you won't be able to
	// render very many of them at the same time
	status = NewGWorld ( &data->effectGWorld, 32, &srcRect, 0, NULL, 0 );
	if ( status != noErr || data->effectGWorld == NULL )
		goto bail;

	dstPixMapHandle = GetGWorldPixMap( data->effectGWorld );
	dstPixFlags = GetPixelsState ( dstPixMapHandle );
	LockPixels( dstPixMapHandle );

	dstRect = srcRect;
	
	// do effect here!
	srcBaseAddr		 = (unsigned char*)GetPixBaseAddr ( srcPixMapHandle );
	srcRowBytes		 = (*srcPixMapHandle)->rowBytes & 0x3FFF;
        srcBaseAddr+=srcRowBytes*srcRect.top+(srcRect.left<<2); 

	dstBaseAddr		 = (unsigned char*)GetPixBaseAddr ( dstPixMapHandle );
	dstRowBytes		 = (*dstPixMapHandle)->rowBytes & 0x3FFF;
        dstBaseAddr+=dstRowBytes*dstRect.top+(dstRect.left<<2); 
	
        rowMax = dstRect.bottom - dstRect.top; 
        colMax = dstRect.right - dstRect.left; 
        
        {long col=colMax; 
     //   long *tp=(long*)(((long)dstBaseAddr)+dstRowBytes); 
        while(col--) {
            ((long*)dstBaseAddr)[col]=0x00000000L; //clear to black
     //       tp[col]=0x00000000L; //clear to black
            }
        }
        
    if (interlaced) {
        
        if (useAltiVec) {
            //these data stream settings aren't right, but iMovie is so slow I can't tell the difference
            vec_dst(srcBaseAddr, 0x02000020, 0); 
            
            vec_dststt(dstBaseAddr, 0x02000020, 3); 
            }
        
        rowMax>>=1; 
        
	for ( row = 0; row < rowMax; row++ ) {
            long col; 
            unsigned long *src0aLineAddr, *src1aLineAddr, *src0bLineAddr, *src1bLineAddr;
            unsigned long *dstaLineAddr, *dstbLineAddr;
            long srcRow0, frac1, frac0; 
		dstaLineAddr = (unsigned long*)(((long)dstBaseAddr)+ (dstRowBytes*row<<1) );
                
                srcRow0= ( (((row<<1)-rowMax)<<18)/3 + (rowMax<<16))>>1; 
                frac1=srcRow0&0x00ffffL; 
                srcRow0>>=16; 
                
                if (interpolation)
                    frac1 = ((frac1>>4)*( ((frac1>>4)*( (0x030000 - 2 * frac1)>>4 )) >> 12)) >> 8; 
                
                frac0=0x0010000L-frac1; 
                
                dstbLineAddr = (unsigned long *)( ((long)dstaLineAddr) + dstRowBytes); 
                
		src0aLineAddr = (srcRow0<0||srcRow0>=rowMax)?(unsigned long*)dstBaseAddr
                    :(unsigned long*)(((long)srcBaseAddr)+ (srcRowBytes*srcRow0<<1) );
                src0bLineAddr = (srcRow0+1<0||srcRow0+1>=rowMax)?(unsigned long*)dstBaseAddr
                    :(unsigned long*)(((long)src0aLineAddr)+ (srcRowBytes) ); 
                src1aLineAddr = (srcRow0+2<0||srcRow0+2>=rowMax)?(unsigned long*)dstBaseAddr
                    :(unsigned long*)(((long)src0aLineAddr)+ (srcRowBytes<<1) ); 
                src1bLineAddr = (srcRow0+3<0||srcRow0+3>=rowMax)?(unsigned long*)dstBaseAddr
                    :(unsigned long*)(((long)src0bLineAddr)+ (srcRowBytes<<1) ); 
                
                if (useAltiVec) {
                    
                    long colMaxV=colMax>>2; 
                    vector unsigned short fracV; 
                    vector unsigned int zero=(vector unsigned int)(0), 
                        shifter=(vector unsigned int)(23); 
                    ((unsigned short*)&fracV)[0]=frac0>>1; //15-bit precision
                    ((unsigned short*)&fracV)[1]=frac1>>1; 
                    fracV = (vector unsigned short)vec_splat((vector unsigned long)fracV, 0); 
                    
                    for ( col = 0; col<colMaxV; col++) {
                        long byteOffset=col<<4; 
                        vector unsigned char src0av, src0bv; 
                        vector unsigned char src1av, src1bv; 
                        vector unsigned char vha, vhb, vla, vlb; 
                        vector unsigned char dstav, dstbv; 
                        
                        src0av = vec_ld(byteOffset, (unsigned char*)src0aLineAddr); 
                        src0bv = vec_ld(byteOffset, (unsigned char*)src0bLineAddr); 
                        src1av = vec_ld(byteOffset, (unsigned char*)src1aLineAddr); 
                        src1bv = vec_ld(byteOffset, (unsigned char*)src1bLineAddr); 
                        
                        vha = vec_mergeh(src0av, src1av); 
                        vhb = vec_mergeh(src0bv, src1bv); 
                        vla = vec_mergel(src0av, src1av); 
                        vlb = vec_mergel(src0bv, src1bv); 
                        
                        dstav = vec_packsu(
                        vec_packsu(
                            vec_sr(vec_msums((vector unsigned short)vec_mergeh(vha, vha), fracV, zero), shifter),
                            vec_sr(vec_msums((vector unsigned short)vec_mergel(vha, vha), fracV, zero), shifter)), 
                        vec_packsu(
                            vec_sr(vec_msums((vector unsigned short)vec_mergeh(vla, vla), fracV, zero), shifter),
                            vec_sr(vec_msums((vector unsigned short)vec_mergel(vla, vla), fracV, zero), shifter))
                            );
                        dstbv = vec_packsu(
                        vec_packsu(
                            vec_sr(vec_msums((vector unsigned short)vec_mergeh(vhb, vhb), fracV, zero), shifter),
                            vec_sr(vec_msums((vector unsigned short)vec_mergel(vhb, vhb), fracV, zero), shifter)), 
                        vec_packsu(
                            vec_sr(vec_msums((vector unsigned short)vec_mergeh(vlb, vlb), fracV, zero), shifter),
                            vec_sr(vec_msums((vector unsigned short)vec_mergel(vlb, vlb), fracV, zero), shifter))
                            );
                        
                        vec_stl(dstav, byteOffset, (vector unsigned char*)dstaLineAddr); 
                        vec_stl(dstbv, byteOffset, (vector unsigned char*)dstbLineAddr); 
                        }
                                    
                    }
                else 
		for ( col = colMax; col--; ) {
                        unsigned long src0a=src0aLineAddr[col]; 
                        unsigned long src0b=src0bLineAddr[col]; 
                        unsigned long src1a=src1aLineAddr[col]; 
                        unsigned long src1b=src1bLineAddr[col]; 
                        unsigned long dsta, dstb; 
                        
                        dsta = 0xff000000L | 
                        (
                        ( ((src0a&0x00ff0000L) >> 16) * frac0
                        + 
                        ((src1a&0x00ff0000L) >> 16) * frac1) & 0x00ff0000L
                        ) 
                        |
                        ( (
                        ( ((src0a&0x00ff00L) >> 8) * frac0
                        + 
                        ((src1a&0x00ff00L) >> 8) * frac1) >> 8
                        ) & 0x0000ff00L )
                        |
                        ( (
                        ( (src0a&0x00ffL) * frac0
                        + 
                        (src1a&0x00ffL) * frac1) >> 16
                        ) & 0x0000ffL ); 
                        
                        dstb = 0xff000000L | 
                        (
                        ( ((src0b&0x00ff0000L) >> 16) * frac0
                        + 
                        ((src1b&0x00ff0000L) >> 16) * frac1) & 0x00ff0000L
                        ) 
                        |
                        ( (
                        ( ((src0b&0x00ff00L) >> 8) * frac0
                        + 
                        ((src1b&0x00ff00L) >> 8) * frac1) >> 8
                        ) & 0x0000ff00L )
                        |
                        ( (
                        ( (src0b&0x00ffL) * frac0
                        + 
                        (src1b&0x00ffL) * frac1) >> 16
                        ) & 0x0000ffL ); 
                        
                        dstaLineAddr[col] = dsta; 
                        dstbLineAddr[col] = dstb; 
    		}
                
			
	}
        
        }
    else {
        
        if (useAltiVec) {
            //these data stream settings aren't right, but iMovie is so slow I can't tell the difference
            vec_dst(srcBaseAddr, 0x02000020, 0); 
            
            vec_dststt(dstBaseAddr, 0x02000020, 3); 
            }
        
	for ( row = 0; row < rowMax; row++ ) {
            long col; 
            unsigned long *src0LineAddr, *src1LineAddr, *dstLineAddr;
            long srcRow0, frac1, frac0; 
		dstLineAddr = (unsigned long*)(((long)dstBaseAddr)+dstRowBytes*row);
                
                srcRow0= ( (((row<<1)-rowMax)<<18)/3 + (rowMax<<16))>>1; 
                frac1=srcRow0&0x00ffffL; 
                srcRow0>>=16; 
                
                if (interpolation)
                    frac1 = ((frac1>>4)*( ((frac1>>4)*( (0x030000 - 2 * frac1)>>4 )) >> 12)) >> 8; 
                
		src0LineAddr = (srcRow0<0||srcRow0>=rowMax)?(unsigned long*)dstBaseAddr
                    :(unsigned long*)(((long)srcBaseAddr)+srcRowBytes*srcRow0);
                src1LineAddr = (srcRow0+1<0||srcRow0+1>=rowMax)?(unsigned long*)dstBaseAddr
                    :(unsigned long*)(((long)src0LineAddr)+srcRowBytes); 
                frac0=0x0010000L-frac1; 
                
                if (useAltiVec) {
                    
                    long colMaxV=colMax>>2; 
                    vector unsigned short fracV; 
                    vector unsigned int zero=(vector unsigned int)(0), 
                        shifter=(vector unsigned int)(23); 
                    ((unsigned short*)&fracV)[0]=frac0>>1; //15-bit precision
                    ((unsigned short*)&fracV)[1]=frac1>>1; 
                    fracV = (vector unsigned short)vec_splat((vector unsigned long)fracV, 0); 
                    
                    for ( col = 0; col<colMaxV; col++) {
                        long byteOffset=col<<4; 
                        vector unsigned char src0v; 
                        vector unsigned char src1v; 
                        vector unsigned char vh, vl; 
                        
                        src0v = vec_ld(byteOffset, (unsigned char*)src0LineAddr); 
                        src1v = vec_ld(byteOffset, (unsigned char*)src1LineAddr); 
                        
                        vh = vec_mergeh(src0v, src1v); 
                        vl = vec_mergel(src0v, src1v); 
                        
                        vec_stl( 
                        vec_packsu(
                        vec_packsu(
                            vec_sr(vec_msums((vector unsigned short)vec_mergeh(vh, vh), fracV, zero), shifter),
                            vec_sr(vec_msums((vector unsigned short)vec_mergel(vh, vh), fracV, zero), shifter)), 
                        vec_packsu(
                            vec_sr(vec_msums((vector unsigned short)vec_mergeh(vl, vl), fracV, zero), shifter),
                            vec_sr(vec_msums((vector unsigned short)vec_mergel(vl, vl), fracV, zero), shifter))
                            ), 
                            byteOffset, (vector unsigned char*)dstLineAddr); 
                        }
                    
                    /* lower precision (7 bits), but possibly faster
                    long colMaxV=colMax>>2; 
                    vector unsigned char frac0v, frac1v; 
                    vector unsigned char myperm=(vector unsigned char)(0x00,0x10, 0x02,0x12, 
                        0x04,0x14, 0x06,0x16, 
                        0x08,0x18, 0x0a,0x1a, 
                        0x0c,0x1c, 0x0e,0x1e); 
                    ((unsigned char*)&frac0v)[0]=frac0>>9; 
                    ((unsigned char*)&frac1v)[0]=frac1>>9; 
                    frac0v = vec_splat(frac0v, 0); 
                    frac1v = vec_splat(frac1v, 0); 
                    
                    for ( col = 0; col<colMaxV; col++) {
                        long byteOffset=col<<4; 
                        vector unsigned char src0v; 
                        vector unsigned char src1v; 
                        vector unsigned char dst; 
                        
                        src0v = vec_ld(byteOffset, (unsigned char*)src0LineAddr); 
                        src1v = vec_ld(byteOffset, (unsigned char*)src1LineAddr); 
                        
                        dst = (vector unsigned char) vec_perm(
                                vec_add(vec_mule(src0v, frac0v), vec_mule(src1v, frac1v)),
                                vec_add(vec_mulo(src0v, frac0v), vec_mulo(src1v, frac1v)),
                                myperm); 
                        vec_stl(vec_add(dst, dst), 
                            byteOffset, (vector unsigned char*)dstLineAddr); 
                        }
                    */
                    }
                else 
		for ( col = colMax; col--; ) {
                        unsigned long src0=src0LineAddr[col]; 
                        unsigned long src1=src1LineAddr[col]; 
                        
                        //linear, for now
                        dstLineAddr[col] = 0xff000000L | 
                        (
                        ( ((src0&0x00ff0000L) >> 16) * frac0
                        + 
                        ((src1&0x00ff0000L) >> 16) * frac1) & 0x00ff0000L
                        ) 
                        |
                        ( (
                        ( ((src0&0x00ff00L) >> 8) * frac0
                        + 
                        ((src1&0x00ff00L) >> 8) * frac1) >> 8
                        ) & 0x0000ff00L )
                        |
                        ( (
                        ( (src0&0x00ffL) * frac0
                        + 
                        (src1&0x00ffL) * frac1) >> 16
                        ) & 0x0000ffL ); 
    		}
			
            }
        }
        
	
	{
        GWorldFlags framePixFlags;
	framePixMapHandle = GetGWorldPixMap( params->frame );
	framePixFlags = GetPixelsState ( framePixMapHandle );
	LockPixels( framePixMapHandle );
	
	// We need to call CopyBits so it can handle the scaling
	CopyBits ( GetPortBitMapForCopyBits((GrafPtr)data->effectGWorld), GetPortBitMapForCopyBits((GrafPtr)params->frame), &srcRect, &params->frameRect, 0, NULL);
bail:	
	if ( srcPixMapHandle )
		SetPixelsState ( srcPixMapHandle, srcPixFlags );
	if ( dstPixMapHandle )
		SetPixelsState ( dstPixMapHandle, dstPixFlags );
	if ( framePixMapHandle )
		SetPixelsState ( framePixMapHandle, framePixFlags );
	}
        
	if ( data->effectGWorld ) {
		DisposeGWorld ( data->effectGWorld );
		data->effectGWorld = NULL;
	}
	
	if ( status != noErr ) {
		if ( status == memFullErr )
			result = kPluginOutOfMemory;
		else
			result = kPluginFailed;
	}
		
	return result;
}

static HYPluginResult EffectDrive ( short which, HYPluginParams *params )
{
	#pragma unused ( which )
	OSErr			status = noErr;
	HYPluginResult	result = kPluginOK;
	EffectData 		*data;
	CGrafPtr		oldPort;
	GDHandle		oldGD;
	Rect			srcRect, dstRect;
	GWorldPtr		gWorldBefore = NULL, gWorldOriginal = NULL;
	short			width, height, volume, opacity = 0;
	long			frame;
	HYRefDVQuality	quality;
	HYFrameCookie 	frameCookie1;
	PixMapHandle 	srcPixMapHandle, dstPixMapHandle;
	char			srcHandleFlags, dstHandleFlags;
	GWorldFlags		srcPixFlags, destPixFlags;
	
	data = (EffectData *)params->clientData;
	if ( !data )
		return kPluginBadParams;
		
	GetGWorld ( &oldPort, &oldGD );
	SetGWorld ( params->frame, NULL );
	
	// Now get a frame to draw effect in
	width = params->frameRect.right - params->frameRect.left;
	height = params->frameRect.bottom - params->frameRect.top;
	
	frame = params->startFrame + params->renderFrame;

	if ( frame < 0 )
		goto bail;
		
	if ( params->proxyOnly && height < 260 )
		quality = kDVLowQuality;
	else
		quality = kDVHighQuality;
	
	// We always need at least one frame
	result = prox->GetFrame ( params->instance, frame, &gWorldBefore, width, height, quality, &volume, &frameCookie1 );
	if ( result != kPluginOK )
		goto bail;
	
	srcRect = params->frameRect;
	OffsetRect ( &srcRect, - srcRect.left, - srcRect.top );
	
	if ( data->fadeInFrames > 0 || data->fadeOutFrames != 0 ) {
		if ( params->renderFrame < data->fadeInFrames || params->renderFrame > params->frameCount - data->fadeOutFrames ) {
                        double percent; 
			status = NewGWorld ( &gWorldOriginal, 32, &srcRect, 0, NULL, 0 );
			if ( status != noErr || gWorldOriginal == NULL )
				goto bail;
			
			// copy gWorldBefore into gWorldOriginal
			srcPixMapHandle = GetGWorldPixMap(gWorldOriginal );
			srcHandleFlags = HGetState ( (Handle)srcPixMapHandle );
			MoveHHi((Handle) srcPixMapHandle);
			HLock((Handle) srcPixMapHandle);
			srcPixFlags = GetPixelsState ( srcPixMapHandle );
			LockPixels(srcPixMapHandle);
			
			CopyBits ( GetPortBitMapForCopyBits((GrafPtr)gWorldBefore), GetPortBitMapForCopyBits((GrafPtr)gWorldOriginal), &srcRect, &srcRect, 0, NULL);
			
			HSetState ( (Handle) srcPixMapHandle, srcHandleFlags );
			SetPixelsState ( srcPixMapHandle, srcPixFlags );
				
			if ( params->renderFrame < data->fadeInFrames ) {
				percent = ((double)params->renderFrame / (double)data->fadeInFrames);
			} else if ( params->renderFrame > params->frameCount - data->fadeOutFrames ) {
				percent = ((double)(params->frameCount - params->renderFrame) / (double)data->fadeOutFrames);
			}
			opacity = (short)(1.0 - percent) * 255;
		}
	}
	
	result = DoSqueezeToLetterboxEffect ( gWorldBefore, params );

	/* Handle Effect Fade In/Out */
	if ( gWorldOriginal ) {
		GetPortBounds((GrafPtr)gWorldBefore, &srcRect);
		
		dstRect = params->frameRect;
		
		srcPixMapHandle = GetGWorldPixMap(gWorldOriginal );
		srcHandleFlags = HGetState ( (Handle)srcPixMapHandle );
		MoveHHi((Handle) srcPixMapHandle);
		HLock((Handle) srcPixMapHandle);
		srcPixFlags = GetPixelsState ( srcPixMapHandle );
		LockPixels(srcPixMapHandle);

		dstPixMapHandle = GetGWorldPixMap ( params->frame );
		dstHandleFlags = HGetState ( (Handle)dstPixMapHandle );
		MoveHHi((Handle) dstPixMapHandle);
		HLock((Handle) dstPixMapHandle);
		destPixFlags = GetPixelsState ( dstPixMapHandle );
		LockPixels(dstPixMapHandle);

		result = prox->Blit ( *srcPixMapHandle, *dstPixMapHandle, &srcRect, &dstRect, opacity );
		if ( result != kPluginOK )
			goto bail;
		
		HSetState ( (Handle) srcPixMapHandle, srcHandleFlags );
		SetPixelsState ( srcPixMapHandle, srcPixFlags );
		HSetState ( (Handle) dstPixMapHandle, dstHandleFlags );
		SetPixelsState ( dstPixMapHandle, destPixFlags );
	}
	
	// Handle the audio
	
	if ( result == kPluginOK ) {
	
		if ( !params->proxyOnly ) {

			result = prox->MixAudio ( params->instance, frameCookie1, (short)254.0 /* percent * volume */ );
                        /* This was 256.  It seems to cause some bothersome noise in the audio, as if the 
                        value wrapped over.  The Adjust Colors that comes with iMovie exhibits the same thing.  */
		}
	}

bail:
		
	HYPluginResult newResult = prox->ReleaseFrame ( params->instance, frameCookie1 );
	if ( result == kPluginOK )
		result = newResult;

	if ( gWorldOriginal ) {
		DisposeGWorld ( gWorldOriginal );
	}
	
	SetGWorld ( oldPort, oldGD );
	
	if ( result == kPluginOK && status != noErr )
		result = kPluginFailed;

	return result;
}

static HYPluginResult EffectEnd ( short which, HYPluginParams *params )
{
	#pragma unused ( which )
	HYPluginClipInfo clipInfo;
	
	prox->GetClipInfo ( params->instance, kThisClip, &clipInfo );
	
	
	if ( params->clientData ) {

		EffectData *data = (EffectData *)(params->clientData);

		if ( data->effectGWorld ) {
			DisposeGWorld ( data->effectGWorld );
			data->effectGWorld = NULL;
		}
		prox->DisposePtr ( (Ptr)params->clientData );
		params->clientData = NULL;
	}
	
	return kPluginOK;
}

// ----------------------------------------------------------------------
#pragma mark -


static HYPluginResult MakeEffectPlaceHolder ( short which, HYPluginParams *params )
{
	HYPluginResult result = kPluginOK;
	short frameWidth, frameHeight;
	HYPluginParams newParams;
	
	CGrafPtr	currPort;
	GDHandle	currGD;
	
	GetGWorld ( &currPort, &currGD );
	
	frameWidth = params->frameRect.right - params->frameRect.left;
	frameHeight = params->frameRect.bottom - params->frameRect.top;

	SetGWorld ( params->frame, NULL );

	newParams = *params;
	newParams.renderFrame = newParams.getInfo.bestThumbOffset;
	newParams.selector = kPlugInitialize;
	result = EffectBegin ( which, &newParams );
	if ( result != kPluginOK )
		goto done;
		
	newParams.selector = kPlugDoFrame;

	EffectDrive ( which, &newParams );

	if ( result != kPluginOK )
		goto done;

	newParams.selector = kPlugTerminate;
	EffectEnd ( which, &newParams );

done:
	SetGWorld ( currPort, currGD );
	
	return result;
}


